@@ -1,24 +1,25 @@
 """
     Unofficial Python API for retrieving data from del.icio.us.
 
-    This module provides the following features:
+    This module provides the following features plus some more:
 
-    * getting a url's history with
-        * common tags (up to a maximum of 25) including weight (0...5)
+    * getting a url's full history with
         * users who bookmarked the url including tags used for such bookmarks
       	  and the creation time of the bookmark in YYYY-MM format
-    * getting a user's tags including tag counts, i.e. her/his tagging vocabulary
+        * common tags (up to a maximum of 25) including weight (0...5)
+    * getting a user's full bookmark collection including private bookmarks
+      IF you know username AND password
+    * getting a user's most recent public bookmarks if you do not know the
+      password
+    * getting a user's full tagging vocabulary, i.e. tags and tag counts,
+      in case you do not know the password
     * HTTP proxy support
 
-    Note: Only public del.icio.us data will be mined (read below). This means
-    that this API does not (yet) provide means to access your private bookmark
-    data.
-
-    The official del.icio.us API does not provide the functionality mentioned
-    above, so this module will query the del.icio.us website directly and
-    extract the required information by parsing the HTML code of the resulting
-    web pages (a kind of poor man's web mining). The module is able to detect
-    IP throttling, which is employed by del.icio.us to temporarily block
+    The official del.icio.us API does not provide all the functionality mentioned
+    above, and in such cases this module will query the del.icio.us *website*
+    directly and extract the required information by parsing the HTML code of the
+    resulting web pages (a kind of poor man's web mining). The module is able to
+    detect IP throttling, which is employed by del.icio.us to temporarily block
     abusive HTTP request behavior, and will raise a custom Python error to
     indicate that. Please be a nice netizen and do not stress the del.icio.us
     service more than necessary.
@@ -37,8 +38,9 @@
 
 """
 
-__version__ = "1.3.3"
+__version__ = "1.4.1"
 
+import base64
 import cgi
 import datetime
 import httplib
@@ -60,6 +62,76 @@ except:
     print ""
     raise
 
+try:
+    import feedparser
+except:
+    print "ERROR: could not import Universal Feed Parser module"
+    print
+    print "You can download Universal Feed Parser from the Python Cheese Shop at"
+    print "http://pypi.python.org/pypi/FeedParser"
+    print
+    raise
+
+
+class DeliciousUser(object):
+    """This class wraps all available information about a user into one object.
+
+    Variables:
+        bookmarks:
+            A list of (user, tags, title, notes, timestamp) tuples representing
+            a user's bookmark collection.
+
+            url is a 'unicode'
+            tags is a 'list' of 'unicode'
+            title is a 'unicode'
+            notes is a 'unicode'
+            timestamp is a 'datetime.datetime'
+
+        tags (read-only property):
+            A list of (tag, tag_count) tuples, aggregated over all a user's
+            retrieved bookmarks. The tags represent a user's tagging vocabulary.
+
+        username:
+            The del.icio.us account name of the user.
+
+    """
+
+    def __init__(self, username, bookmarks=None):
+        assert username
+        self.username = username
+        self.bookmarks = bookmarks or []
+
+    def __str__(self):
+        total_tag_count = 0
+        total_tags = set()
+        for url, tags, title, notes, timestamp in self.bookmarks:
+            total_tag_count += len(tags)
+            for tag in tags:
+                total_tags.add(tag)
+        return "[%s] %d bookmarks, %d tags (%d unique)" % \
+                    (self.username, len(self.bookmarks), total_tag_count, len(total_tags))
+
+    def __repr__(self):
+        return self.username
+
+    def get_tags(self):
+        """Returns a dictionary mapping tags to their tag count.
+
+        For example, if the tag count of tag 'foo' is 23, then
+        23 bookmarks were annotated with 'foo'. A different way
+        to put it is that 23 users used the tag 'foo' when
+        bookmarking the URL.
+
+        """
+
+        total_tags = {}
+        for url, tags, title, notes, timestamp in self.bookmarks:
+            for tag in tags:
+                total_tags[tag] = total_tags.get(tag, 0) + 1 
+        return total_tags
+    tags = property(fget=get_tags, doc="Returns a dictionary mapping tags to their tag count")
+
+
 class DeliciousURL(object):
     """This class wraps all available information about a web document into one object.
 
@@ -71,17 +143,17 @@ class DeliciousURL(object):
             user is a 'str', tags is a 'list', timestamp is a 'str' in
             'YYYY-MM' format
 
+        tags (read-only property):
+            A list of (tag, tag_count) tuples, aggregated over all a document's
+            retrieved bookmarks.
+
         common_tags:
             A list of (tag, tag_weight) tuples, representing a document's so-called
             "common tags", i.e. the up to 25 most popular tags for this document.
 
-            tag is a 'str', tag_weight is an 'int'
-
         url:
             The URL of the document.
 
-            url is a 'str'
-
     """
 
     def __init__(self, url, common_tags=None, bookmarks=None):
@@ -97,7 +169,7 @@ class DeliciousURL(object):
             total_tag_count += len(tags)
             for tag in tags:
                 total_tags.add(tag)
-        return "[%s] %d users (= bookmarks), %d tags (%d unique), %d out of 25 max 'common' tags" % \
+        return "[%s] %d bookmarks (= users), %d tags (%d unique), %d out of 25 max 'common' tags" % \
                     (self.url, len(self.bookmarks), total_tag_count, \
                     len(total_tags), len(self.common_tags))
 
@@ -176,7 +248,7 @@ class DeliciousAPI(object):
         socket.setdefaulttimeout(self.timeout)
 
 
-    def _query(self, path):
+    def _query(self, path, host="del.icio.us", user=None, password=None, use_ssl=False):
         """Queries delicious for information, specified by (query) path.
 
         Returns None on errors (i.e. on all HTTP status other than 200).
@@ -187,21 +259,35 @@ class DeliciousAPI(object):
         # compose delicious query
         method = "GET"
         headers = { 'User-agent' : self.user_agent }
+        # add HTTP Basic authentication if available
+        if user and password:
+            encodedCredentials = base64.encodestring('%s:%s' % (user, password))[:-1]
+            headers["Authorization"] = "Basic %s" % encodedCredentials
+
         body = None
 
         conn = None
+        if use_ssl:
+            port = 443
+        else:
+            port = 80
         if self.http_proxy:
             host, port = self.http_proxy.split(":")
-            conn = httplib.HTTPConnection(host, port)
+        if use_ssl:
+            conn = httplib.HTTPSConnection(host, port)
         else:
-            conn = httplib.HTTPConnection("del.icio.us")
+            conn = httplib.HTTPConnection(host, port)
 
         data = None
 
         tries = self.tries
+        if use_ssl:
+            protocol = "https"
+        else:
+            protocol = "http"
         while tries > 0:
             try:
-                conn.request(method, "".join(["http://del.icio.us", path]), body, headers)
+                conn.request(method, "".join([protocol, "://", host, path]), body, headers)
                 # read response from server
                 response = conn.getresponse()
 
@@ -209,6 +295,8 @@ class DeliciousAPI(object):
                     data = response.read()
                 elif response.status == 302:
                     raise DeliciousMovedTemporarilyWarning, "del.icio.us status %s - url moved temporarily" % response.status
+                elif response.status == 401:
+                    raise DeliciousUnauthorizedError, "del.icio.us error %s - unauthorized (authentication failed?)" % response.status
                 elif response.status == 404:
                     raise DeliciousNotFoundError, "del.icio.us error %s - url not found" % response.status
                 elif response.status == 500:
@@ -252,7 +340,7 @@ class DeliciousAPI(object):
         return data
 
     def get_url(self, url):
-        """Returns a DeliciousURL representation of the del.icio.us history of url.
+        """Returns a DeliciousURL instance representing the del.icio.us history of url.
 
         Generally, this method is what you want for getting bookmark, tag, and
         user information about a URL.
@@ -316,6 +404,20 @@ class DeliciousAPI(object):
         bookmarks.reverse()
         return bookmarks
 
+    def get_common_tags_of_url(self, url):
+        """Returns the list of "common tags" of url from del.icio.us."""
+        tags = []
+
+        m = md5.new()
+        m.update(url)
+        hash = m.hexdigest()
+
+        path = "".join(["/url/", hash])
+        data = self._query(path)
+        if data:
+            tags = self._extract_common_tags_from_url_history(data)
+        return tags
+
     def _extract_common_tags_from_url_history(self, data):
         common_tags = []
         soup = BeautifulSoup(data)
@@ -334,24 +436,81 @@ class DeliciousAPI(object):
                 common_tags.append( (tag, weight) )
         return common_tags
 
-    def get_common_tags_of_url(self, url):
-        """Returns the list of "common tags" of url from del.icio.us."""
-        tags = []
+    def get_user(self, username, password=None):
+        """Retrieves a user's bookmarks from del.icio.us.
 
-        m = md5.new()
-        m.update(url)
-        hash = m.hexdigest()
+        If a correct username AND password are supplied, a user's *full*
+        bookmark collection (which also includes private bookmarks) is
+        retrieved. Data communication is encrypted using SSL in this case.
 
-        path = "".join(["/url/", hash])
-        data = self._query(path)
-        if data:
-            tags = self._extract_common_tags_from_url_history(data)
-        return tags
+        If no password is supplied, only the most recent public bookmarks
+        of the user are extracted from his/her RSS feed (about 30 bookmarks
+        if any). Note that if you want to get the *full* tagging vocabulary
+        of the user even if you don't know the password, you can call
+        get_tags_of_user() instead.
+
+        This function can be used to backup all of a user's bookmarks if
+        called with a username and password.
+
+        Returns a DeliciousUser instance.
+
+        """
+        assert username
+        user = DeliciousUser(username)
+        bookmarks = []
+        if password:
+            # We have username AND password, so we call
+            # the official del.ici.us API.
+            path = "/v1/posts/all"
+            data = self._query(path, host="api.del.icio.us", use_ssl=True, user=username, password=password)
+            if data:
+                soup = BeautifulSoup(data)
+                elements = soup.findAll("post")
+                for element in elements:
+                    url = element["href"]
+                    title = element["description"] or u""
+                    notes = element["extended"] or u""
+                    tags = None
+                    if element["tag"]:
+                        tags = element["tag"].split()
+                    timestamp = datetime.datetime.strptime(element["time"], "%Y-%m-%dT%H:%M:%SZ")
+                    bookmarks.append( (url, tags, title, notes, timestamp) )
+            user.bookmarks = bookmarks
+        else:
+            # We have only the username, so we extract data from
+            # the user's RSS feed. However, the feed is restricted
+            # to the most recent public bookmarks of the user (about 30
+            # if any)
+            path = "/rss/%s" % username
+            data = self._query(path, host="feeds.delicious.com", user=username)
+            if data:
+                f = feedparser.parse(data)
+                for entry in f.entries:
+                    url = entry.link
+                    tags = []
+                    if entry.tags and "term" in entry.tags[0]:
+                        tags = entry.tags[0].term.split()
+                    else:
+                        tags.append(u"system:unfiled")
+                    title = entry.title
+                    notes = entry.summary
+                    timestamp = datetime.datetime.strptime(entry.updated, "%Y-%m-%dT%H:%M:%SZ")
+                    bookmarks.append( (url, tags, title, notes, timestamp) )
+                user.bookmarks = bookmarks
+        return user
 
     def get_tags_of_user(self, username):
-        """Returns user's tags from del.icio.us (his tagging 'vocabulary', as used in displaying the tag cloud)."""
-        tags = []
+        """
+        Retrieves user's tags and their tag counts from del.icio.us.
+        The tags represent a user's full tagging vocabulary.
 
+        IF you know the user's password, then get_user() provides
+        even more information.
+
+        Returns a dictionary mapping tags to their tag counts.
+
+        """
+        tags = {}
         path = "".join(["/", username, "?setminposts=1&setbundleview=hide"])
         data = self._query(path)
         if data:
@@ -386,7 +545,7 @@ class DeliciousAPI(object):
 
             assert len(counts) == len(tags)
             tags_counts = zip(tags, counts)
-        return tags_counts
+        return dict(tags_counts)
 
     def get_number_of_users(self, url):
         """Returns the number of users who have rated this url at del.icio.us."""
@@ -463,6 +622,15 @@ class DeliciousUnknownError(DeliciousError):
     """Used to indicate that del.icio.us returned an (HTTP) error which we don't know how to handle yet."""
     pass
 
+class DeliciousUnauthorizedError(DeliciousError):
+    """Used to indicate that del.icio.us returned a 401 Unauthorized error.
+
+    Most of the time, the user credentials for acessing restricted (official)
+    del.icio.us API functions are incorrect.
+
+    """
+    pass
+
 class DeliciousNotFoundError(DeliciousError):
     """Used to indicate that del.icio.us returned a 404 Not Found error.
 
@@ -485,7 +653,7 @@ class DeliciousMovedTemporarilyWarning(DeliciousWarning):
     """Used to indicate that del.icio.us returned a 302 Found (Moved Temporarily) redirection."""
     pass
 
-__all__ = ['DeliciousAPI', 'DeliciousURL', 'DeliciousError', 'DeliciousThrottleError', 'DeliciousUnknownError', 'DeliciousNotFoundError' , 'Delicious500Error', 'DeliciousMovedTemporarilyWarning']
+__all__ = ['DeliciousAPI', 'DeliciousURL', 'DeliciousError', 'DeliciousThrottleError', 'DeliciousUnauthorizedError', 'DeliciousUnknownError', 'DeliciousNotFoundError' , 'Delicious500Error', 'DeliciousMovedTemporarilyWarning']
 
 if __name__ == "__main__":
     d = DeliciousAPI()
